#include "Certificate_order.h"
#include "Node.h"

template <class T> 
Certificate_order<T>::Certificate_order(int comp, int corr, int maxs) : bmap(Max_num_replicas) {
  max_size = (maxs == 0) ? node->f()*5+1 : maxs;
  vals = new Message_val_order[max_size];
  all_vals = new Message_val_order[max_size];
  all_id_list = new int(max_size);
  next_elem = 0;

  for (int i = 0; i<max_size; i++) {
    vals[i].id_list = (int *) malloc(max_size*sizeof(int));
    vals[i].reply_list = (T **) malloc(max_size*sizeof(T *));
    for (int j=0; j<max_size;j++) {
      vals[i].reply_list[j] = NULL;
      vals[i].id_list[j] = -1;
    }
  }
  
  cur_size = 0;
  correct = (corr == 0) ? node->f()*4+1 : corr;
  complete = (comp == 0) ? node->f()*4+1 : comp;
  c = 0;
  mym = 0; 
}

template <class T> 
Certificate_order<T>::~Certificate_order() {
  // delete all received order-request messages first
  for (int i =0;i < next_elem; i++) {
      delete all_vals[i].m;
  } 
  delete [] all_vals;
  delete all_id_list;

  // delete all unique message array
  delete [] vals;
}

template <class T> 
bool Certificate_order<T>::add(T *m) {
  const int id = m->id();
  if (node->is_replica(id) && !bmap.test(id)) {
    // If we did not receive from this id
    // if ((c==0)&&(m->verify())) {
    // fprintf(stderr,"SUCCESS in bmap \n"); 
    if ((m->verify())) {
      // add "m" to the certificate	
      // th_assert(id != node->id(), "verify should return false for messages from self");
      //fprintf(stderr,"SUCCESS in Certificate order add \n"); 
      // Update all list 
      bmap.set(id);	
      all_id_list[next_elem] = id;
      
      all_vals[next_elem].m = m;
      //fprintf(stderr,"Adding %x with id : %d seq no : %d in all vals \n", m,m->seqno(),id);

      next_elem++;

      // Check if there is a value that matches "m"
      int i;
      for (i=0; i < cur_size; i++) {
	Message_val_order &val = vals[i];
	if (val.m->match(m)) {
	  val.id_list[val.count] = m->rep_id();
      
	  val.reply_list[val.count] = m;
	  val.count++;
          //fprintf(stderr,"Adding matching %x with id : %d seq no : %d in vals count : %d \n", m,m->seqno(),id,val.count);

          if ((c==0) && (val.count >= correct)) { 
	    c = vals+i;
	    c->lc_last_checked = 0;
	  }
	  update_locally_committed();

               
         if (!val.m->full()) {
           if (m->full()) {
             // if val.m is not full and m is, replace val.m
             val.m = m;
           }
         }
	return true;
        }
      }
      
      // "m" has a new value.
      if (cur_size < max_size) {
	if (m->tag() == Reply_tag) 
	vals[cur_size].m = m;
	vals[cur_size].id_list[0] = m->rep_id();
	vals[cur_size].reply_list[0] = m;
	vals[cur_size].count = 1;
	if (correct == 1) { 
	  // This condition will be true for f = 0; 
          c = &vals[cur_size];
        }
        cur_size++;
        return true;
      } 
    }
  } 
  delete m;
  return false;
}

template <class T> 
void Certificate_order<T>::update_locally_committed() {
   if (c != 0) {
       int cur = c->lc_last_checked;
       for (int i = cur; i<c->count; i++) { 
          if (c->reply_list[i]->check_locally_committed()) c->lc_count++;
       }
       c->lc_last_checked=c->count;
   }	
}	

template <class T> 
bool Certificate_order<T>::collect_proof(struct iovec *&p, int &len, int &size) {
 len =0;
 size =0;
 int nelem=0;

  if ((c != (Message_val_order *)NULL)) {
     // Certificate is complete
     len = c->count+2;
     p = (struct iovec *)malloc(len*(sizeof(struct iovec)));
     
     // 1. Store the matching OR request message
     p[0].iov_base = c->m->read_order();
     p[0].iov_len  = c->m->order_size(); 
     int cur = 1;

     // 2. Now append all the authenticators from each matching replica
     for (int i=0; i<c->count; i++) {       
       // Point to the authenticators from all the replicas
       p[i+cur].iov_base = c->reply_list[i]->read_order_auth();
       p[i+cur].iov_len  = node->auth_size_rep(); 
       size += p[i+cur].iov_len;	
     }
    
     // 3. Now, include the id list of the replicas that sent the matching speculative reply messages
     p[len-1].iov_base = c->id_list;
     p[len-1].iov_len = c->count*sizeof(int);	
     size += p[len-1].iov_len;
     return true;
  }
  else {
   // Certificate is not complete 
   // Client has to include atleast f+1 verifiable certificates
   len = next_elem+1;
   nelem = next_elem;
   p = (struct iovec *)malloc(len*(sizeof(struct iovec)));	

   int cur = 0;
   // 1. Relay all order request messages along with authenticators
   // to prove that they don't match
   for (int i=0; i<next_elem; i++) {
       // Points to the OR message relayed by  this replica 
       p[i+cur].iov_base = all_vals[i].m->read_order();
       p[i+cur].iov_len = all_vals[i].m->order_size()+node->auth_size_rep();
       size += p[i+cur].iov_len;
   }
   cur += next_elem;

   // 2. Now, include the id list of the replicas that sent the matching speculative reply messages
   p[len-1].iov_base = all_id_list;
   p[len-1].iov_len = next_elem*sizeof(int);	
   size += p[len-1].iov_len;
   return false;
   } 
}
        
template <class T> 
T * Certificate_order<T>::cvalue_clear() {
  if (c == 0) {
    return 0;
  }

  T *ret = c->m;
  c->m = 0;
  for (int i=0; i < cur_size; i++) {
    if (vals[i].m == ret)
      vals[i].m = 0;
  }
  clear();

  return ret;
}


