#include <strings.h>
#include "th_assert.h"
#include "Message_tags.h"
#include "Committed_locally.h"
#include "Node.h"
#include "Principal.h"

#include "Statistics.h"

Committed_locally::Committed_locally(Committed_locally_rep *r) : Message(r) {}


Committed_locally::Committed_locally(View view, Request_id req, int replica, Digest &rh_d,
	     Principal *p) : Message(Committed_locally_tag, sizeof(Committed_locally_rep)+MAC_size) {
    rep().v = view; 
    rep().rid = req;
    rep().replica = replica;
    rep().rh_digest = rh_d;

    INCR_OP(reply_auth);
    START_CC(reply_auth_cycles);
    p->gen_mac_out(contents(), sizeof(Committed_locally_rep), contents()+sizeof(Committed_locally_rep));
    STOP_CC(reply_auth_cycles);
    set_size(sizeof(Committed_locally_rep)+MAC_size);
}


void Committed_locally::authenticate(Principal *clientp) {
  int old_size = sizeof(Committed_locally_rep);

  INCR_OP(reply_auth);
  START_CC(reply_auth_cycles);
  clientp->gen_mac_out(contents(), old_size, contents()+old_size);
  STOP_CC(reply_auth_cycles);
  set_size(old_size+MAC_size);
}


bool Committed_locally::verify() {

  // Check signature.
  Principal *replica = node->i_to_p(rep().replica);
  
  INCR_OP(reply_auth_ver);
  START_CC(reply_auth_ver_cycles);

  bool ret = replica->verify_mac_in(contents(), sizeof(Committed_locally_rep), contents()+sizeof(Committed_locally_rep));

  STOP_CC(reply_auth_ver_cycles);

  return ret;
}


bool Committed_locally::convert(Message *m1, Committed_locally *&m2) {
  if (!m1->has_tag(Committed_locally_tag, sizeof(Committed_locally_rep)))
    return false;
  
  m1->trim();
  m2 = (Committed_locally*)m1;
  return true;
}
