#ifndef _Hate_primary_h
#define _Hate_primary_h 1

#include "types.h"
#include "Digest.h"
#include "Message.h"
class Principal;

// 
// Hate_primary messages have the following format:
//
struct Hate_primary_rep : public Message_rep {
  View v;
  int id;         // id of the replica that generated the message.
  // Followed by a signature.
};

class Hate_primary : public Message {
  // 
  //  Hate_primary messages
  //
public:
  Hate_primary(View v);
  // Effects: Creates a new signed Hate_primary message with sequence
  // number "s" and digest "d". "stable" should be true iff the checkpoint
  // is known to be stable.

  View view() const;
  // Effects: Fetches the sequence number from the message.

  int id() const;
  // Effects: Fetches the identifier of the replica from the message.

  bool match(const Hate_primary *c) const;
  // Effects: Returns true iff "c" and "this" have the same digest

  bool verify();
  // Effects: Verifies if the message is signed by the replica rep().id.

  static bool convert(Message *m1, Hate_primary *&m2);
  // Effects: If "m1" has the right size and tag of a "Hate_primary",
  // casts "m1" to a "Hate_primary" pointer, returns the pointer in
  // "m2" and returns true. Otherwise, it returns false. Convert also
  // trims any surplus storage from "m1" when the conversion is
  // successfull.
 
private:
  Hate_primary_rep& rep() const;
  // Effects: Casts "msg" to a Hate_primary_rep&
};

inline Hate_primary_rep& Hate_primary::rep() const { 
  th_assert(ALIGNED(msg), "Improperly aligned pointer");
  return *((Hate_primary_rep*)msg); 
}

inline View Hate_primary::view() const { return rep().v; }

inline int Hate_primary::id() const { return rep().id; }

inline bool Hate_primary::match(const Hate_primary *c) const { 
  return (view() == c->view()); 
}

#endif // _Hate_primary_h
