#include "NV_info.h"
#include "Replica.h"
#include "View_change.h"
#include "View_change_ack.h"
#include "New_view.h"
#include "View_info.h"
#include "Pre_prepare.h"
#include "Status.h"
#include "Array.t"
#include "K_max.h"


//
// NV_info::VC_info methods:
//
  
//
// NV_info methods:
//

NV_info::VC_info::VC_info() : vc(0), ack_count(0), ack_reps(node->n()), req_sum(false) {}

NV_info::VC_info::~VC_info() { delete vc; }

void NV_info::VC_info::clear() {
  delete vc;
  vc = 0;
  ack_count = 0;
  ack_reps.clear();
  req_sum = false;
}

NV_info::Req_sum::Req_sum() : r_proofs(node->n()) {} 

NV_info::Req_sum::~Req_sum() {}

NV_info::NV_info() : v(0), nv(0), vc_cur(0), vc_target(0), vcs(node->n()), vcid(node->n(),false) { 
  vcs._enlarge_by(node->n()); 
  chosen_last_stable = -1;
  chosen_max_committed = -1;
  last_complete = -1;
  done_last_stable = false;
  done_max_committed = false;
  chosen_max_executed = -1;
  max = -1;
  base = -1;
  min = -1;
  n_complete = 0;
  is_complete = false;
  vc_cur =0;
  vack_count = 0;

  nv_sent = zeroTime();
}

NV_info::~NV_info() { clear(); }

void NV_info::clear() {
  v = 0;
  delete nv;
  nv = 0;
  vc_target = 0;

  for (int i=0; i < vcs.size(); i++) {
    vcs[i].clear();
  }
  vc_cur = 0;

  chosen_last_stable = -1;
  chosen_max_committed = -1;
  last_complete = -1;
  done_last_stable = false;
  done_max_committed = false;
  chosen_max_executed = -1;
  max = -1;
  base = -1;
  min = -1;
  nv_sent = zeroTime();
  vack_count =0;

  n_complete = 0;
  is_complete = false;
}

void NV_info::make_complete(View vi) {
  v = vi;
  is_complete = true;
}

bool NV_info::add(New_view* m, View_info* parent) {
  th_assert(m->verify() || node->id() == node->primary(m->view()), "Invalid argument");
  th_assert(parent != 0, "Invalid argument");

  if (m->view() <= v) {
    delete m;
    return false;
  }

  // Remove any old information.
  if (v != 0)
    clear();
  
  // Add m to this.
  v = m->view();
  nv = m;
  vi = parent;

  // Set vc_target.
  for (int i=0; i < node->n(); i++) {
    Digest vd;
    if (m->view_change(i, vd))
      vc_target++;
  }
  
  return true;
}  

bool NV_info::fetch_history_digest(Seqno n, Digest &rd) {
  if (n < nv->min() || n >= nv->max()) {
    return false;
  }
  else {
    Digest d;
    vcs[nv->which_picked(n)].vc->req(n,d, rd);
    return true;
  }
}


bool NV_info::add(View_change *m, bool verified) {
  th_assert(m->view() == v, "Invalid argument");
  
  
  int vid = m->id();
  if (vcs[vid].vc != 0 || is_complete) {
    return false;
  }

  bool is_primary = node->primary(v) == node->id();

  if (!verified) {
    if (is_primary || !m->verify_digest()) {
      return false;
    }
  }
 
  vcs[vid].vc = m;
  vcid.set(m->id());
  vc_cur++;

  
  if ((is_primary)&& m->id() == node->id()) {
    nv->add_view_change(m->id(), m->digest());
    vack_count++;  
  } 

  if (!is_primary && (vc_cur >= vc_target) && (vc_target >= node->n_f())) {
    if(check_new_view()) {
      fprintf(stderr, "Check new view \n");
      replica->process_new_view(min, nv->max_committed(),max);
    }
  }

  return true;
}

bool NV_info::add(View_change_ack* m) {
  th_assert(m->verify() && m->view() == v, "Invalid argument");

  int vci = m->vc_id();
  int mid = m->id();

  bool is_primary = node->primary(v) == node->id();

  if (is_complete)
    return false;

  Digest d;
  bool in_nv = nv->view_change(vci, d);
  
  if (!is_primary) {
    if (!in_nv) 
      return false;
  } else {
    if (!vcs[vci].vc) 
      return false;
 
    d = vcs[vci].vc->digest();
  }

  if (!in_nv && m->vc_digest() == d && !vcs[vci].ack_reps.test(mid)) {
    vcs[vci].ack_reps.set(mid);
    vcs[vci].ack_count++;

    if (vcs[vci].ack_count == node->f()*2-1 && node->primary(v) == node->id()) {
      fprintf(stderr,"ACtually adding vc for : %d \n",vci);
      // This view change has enough acks: add it to the new-view.
      View_change *vc = vcs[vci].vc;
      nv->add_view_change(vci, vc->digest());
      vack_count++;
    }
  }
  
  if (vack_count >= node->n_f()) {
    //fprintf(stderr, "About to create new view at the primary\n");
    summarize();
    make_new_view();
  }   

  delete m;
  return true;
}


void NV_info::summarize() {
  
  Seqno cur_ls = 0;
  Seqno cur_lc = 0;
  // Update last stable and max locally committed 
  for (int i=0; i < vcs.size(); i++) {
  
    if (!nv->view_change(i)) continue; 
  
    cur_ls = chosen_last_stable;
    cur_lc = chosen_max_committed;
 
    // Extract max last stable checkpoint
    View_change *vc = vcs[i].vc;
    int vid = vc->id();
    Seqno ls = vc->last_stable();
    Seqno lc = vc->last_committed();
            
    // Read digest of the last stable checkpoint
    Digest d;
    th_assert(vc->ckpt(d,ls),"No checkpoint in the view change \n");
    //fprintf(stderr, "Summarizing %d vc width digest:", vcs[i].vc->id());
    //d.print(); 
  
    int mcount = 0;
    int mtcount = 0;
    int ccount = 0;
    int ctcount = 0;
    Seqno max_le1 = chosen_max_executed;
    Seqno max_le = chosen_max_executed;

    if ((cur_ls < ls) || (cur_lc < lc)) {
      for (int j=0; j < vcs.size(); j++) {

        if (!nv->view_change(j)) continue; 
	View_change *cvc = vcs[j].vc;
	Seqno cls = cvc->last_stable();
	Seqno clc = cvc->last_committed();
	Seqno cle = cvc->last_executed();
        //fprintf(stderr,"id : %d cls: %llu clc: %llu cle: %llu cur ls : %llu cur lc: %llu max exec: %llu \n", cvc->id(),cls,clc,cle,cur_ls,cur_lc,max_le);  
	// Collect new last stable statistics
	if (cur_ls < ls) {
	  Digest cd;
	  Digest rd;	  
	  Digest rhd;
	  
          if ((ls == cls) && cvc->ckpt(cd,ls) && (d == cd)) {
	    mcount++;
	    mtcount++;
	    if (cle > max_le1) {
	      max_le1 = cle;
	    }
	  }
	  else if ((ls > cls) && (ls <= clc) && cvc->req(ls,rd,rhd) && (d == rhd)) {
	    // Read the committed request history digest
	    // Extract the request history digest
	    mcount++;
	    mtcount++;
	    if (cle > max_le1) {
		max_le1 = cle;
	    }
	  }
	  else if ((ls > clc) && (ls <= cle) && cvc->req(ls,rd,rhd)) {
	    // Extract the request history digest
	    if (d == rhd) {
	      mtcount++;
	      if (cle > max_le1) {
		max_le1 = cle;
	      }
	    }
	  }
	  else if (ls > cle) {
	    mtcount++;
	  }
	  else {}
          //fprintf(stderr, " Iter : %d mcount : %d mtcount : %d ls : %llu cls:%llu  Digest: ", j, mcount, mtcount, ls);
          //cd.print();
	}
	
	// Collect new last committed statistics
	if (cur_lc < lc) {
	  
	  // Extract the request history digest
	  Digest rd;	  
	  Digest rhd;

	  // Read the committed request history digest
	  if ((lc <= clc) && (lc > cls) && cvc->req(lc,rd,rhd) && (d == rhd)) {
	    ccount++;
	    ctcount++;
	    if (cle > max_le) {
	      max_le = cle;
	    }
	  }
	  else if ((lc <= cle) && (lc > cls) && cvc->req(lc,rd,rhd) && (d == rhd)) {
	    ctcount++;
	    if (cle > max_le) {
	      max_le = cle;
	    }
	  }
	  else if (lc > cle) {
	    ctcount++;
	  }
	}
      }	
    }
      
    // Update max_locally_committed
    if ((ccount >= node->f()+1) && (ctcount >= node->n_f()) && (lc >cur_lc)) {
      chosen_max_committed = lc;
      chosen_committed_id = vid;
      done_max_committed = true;
      chosen_max_executed = max_le;
    }
   
    //fprintf(stderr, " mcount : %d mtcount : %d ls : %llu \n", mcount, mtcount, ls);
    // Update last stable
    if ((mcount >= node->f()+1) && (mtcount >=  node->n_f()) && (ls > cur_ls)) {
      chosen_last_stable = ls;
      chosen_last_stable_id = vid;
      done_last_stable = true;
      if (chosen_max_committed < chosen_last_stable) {
	chosen_max_committed = chosen_last_stable;
	chosen_committed_id = vid;
	done_max_committed = true;
	chosen_max_executed = max_le1;
      }
    }
  }

  //fprintf(stderr,"done ls : %u done lc : %u  \n",done_last_stable, done_max_committed);
  // SUMMARIZE REQUESTS
  if ((done_last_stable) && (done_max_committed)) { 
    int cmax = max;
    int cbase = base; 
    min = chosen_last_stable;
    max = chosen_max_executed+1;
    base = chosen_max_committed+1;
    last_complete = min;
    //fprintf(stderr,"min: %llu max : %llu commit : %llu complete: %llu \n",
    //min, max,chosen_max_committed,last_complete);
    
    if ((cmax < max) || (base < cbase)) {
      // Adjust the size of the arrays accordingly
      reqs._enlarge_by(max-base);
      comp_reqs._enlarge_by(max-base);
    }
    
    for (int i=0; i < comp_reqs.size(); i++) {
	comp_reqs[i] = -1;
    }

    for (int j=0; j < vcs.size(); j++) {
      if (nv->view_change(j)) {
	summarize_reqs(vcs[j].vc);
      }
    }
    check_complete();
  }
  
}

void NV_info::check_complete() {
  // Check for holes
  Seqno i = -1;

  // Find the last request with complete proof
  for (i = chosen_max_committed+1; i < max && comp_reqs[i-base] >= 0; i++); 
  last_complete = i-1;
  
}
      

// Pick the request for each sequence number  
// (1) If a request is present in at least f+1 VCs
//      - if there are more than one requests with f+1 matching VCs then 
//        pick any request (it does not matter).
// (2) else if they are present in fewer than f+1 VCs
//      - Pick request that is present in the view change message of 
//       of node with least 

void NV_info::summarize_reqs(View_change *vc) {
  th_assert(vc != 0 && nv->view_change(vc->id()), "Invalid vc argument");
  th_assert(!vcs[vc->id()].req_sum, "Invalid argument");
  th_assert(!is_complete, "Invalid state");

  //fprintf(stderr,"SUMMARIZE reqs min: %llu max : %llu commit : %llu complete: %llu \n",
  //min, max,chosen_max_committed,last_complete);
  Seqno le = vc->last_executed();
  Seqno i = (chosen_max_committed > vc->last_stable()) ? chosen_max_committed : vc->last_stable();
  
  for (i = i+1; i < max; i++) { 
    Array<Req_sum>& reqsi = reqs[i-base];
    bool match = false;
    View rv = -1;

    // Read the committed request history digest
    Digest rd;	  
    Digest rhd;
    if (i <= le) {
      // Extract the request history digest
      rv = vc->request(i,rd,rhd);
    }

    for (int j=0; j < reqsi.size(); j++) {    
      Req_sum& cur = reqsi[j];
      
      if (cur.v == rv && cur.rd == rhd) {
	match = true;
	cur.n_proofs++;
	cur.r_proofs.set(vc->id());
	cur.n_le++;
      } 
      else {
	if ((i > le)) {
	  cur.n_le++;
	}
      }
      
      if (comp_reqs[i-base] < 0 && cur.n_proofs >= node->f()+1 && cur.n_le >= node->n_f()) {
	comp_reqs[i-base] =j;
        //fprintf(stderr,"Req: %d proof: %d le: %d \n",i,cur.n_proofs,cur.n_le); 
	n_complete++;
      }
    }

    if (!match) {
      // No entry matches this request: add a new one.
      reqsi._enlarge_by(1);
      Req_sum& cur = reqsi.high();
      cur.v = rv;
      cur.rd = rhd;
      cur.d = rd;
      cur.n_proofs = 1;
      cur.r_proofs.clear();
      cur.r_proofs.set(vc->id());
      cur.id = vc->id();
      cur.n_le = 1;
 
      // Search view-changes for proofs
      for (int j=0; j < vcs.size(); j++) {
	if (vcs[j].req_sum) {
	  Seqno nle = vcs[j].vc->last_executed(); 
	  if (i > nle) {
	    cur.n_le++;
	  }
	}
      }
    }
  }  	  
   // Set summarize flag to true for this vc
  vcs[vc->id()].req_sum = true;
}

  
void NV_info::choose_ckpt(int index) {
}

void NV_info::make_new_view() {
  th_assert(node->primary(v) == node->id(), "Invalid state");
  th_assert(nv_sent == zeroTime(), "Invalid state");

  nv->set_min(min);
  nv->set_max(max);
  nv->set_complete(last_complete);  
  nv->set_committed(chosen_max_committed);

  fprintf(stderr,"New view: Min : %llu Max : %llu Committed : %llu last complete : %llu \n",
          min, max, chosen_max_committed, last_complete);

  // Pick the checkpoint
  nv->pick(chosen_last_stable_id, min);

  // Pick the requests
  for (Seqno i=min+1; i < chosen_max_committed; i++) {
    nv->pick(chosen_committed_id, i);
  }

  for (Seqno i=chosen_max_committed+1; i <= last_complete; i++) {
     th_assert(comp_reqs[i-base] >= 0, "Invalid state");
     Req_sum& cur = reqs[i-base][comp_reqs[i-base]];
     nv->pick(cur.id, i);
  }

  for (Seqno i=last_complete+1; i <max; i++) {
    nv->pick(node->id(), i);
  }

  is_complete = true;
  nv->re_authenticate();
  nv_sent = currentTime();

  replica->process_new_view(min,nv->max_committed(),max);
}

bool NV_info::check_new_view() {
  th_assert(node->primary(v) != node->id(), "Invalid state");


  // 1. Check if I received all the view change messages in new view message
  int nvcount =0;
  int mvcount =0;
  for (int i=0; i<node->n(); i++) {
    Digest d;
    if (nv->view_change(i,d)) {
      nvcount++;
      if (vcid.test(i)) {
	if (d == vcs[i].vc->digest()) {
	  mvcount++;
	}
	else {
	  // TO DO: I have a different view change message frm node i compared to
	  // what new view message has. This can happen in the presence of a
	  // malicious backup i or new primary. 
	  // Send status message to other replicas to check if they
	  // received the same view change message as the new primary 
	  // has received. If I receive f+1 acks from different 
	  // replicas with same digest d then I know that primary is not 
	  // faulty and hence include the view change message in new view.
	  // Else primary is faulty.
	}
      }
      else {
	// TO DO: New primary received view change message from node i
	// but I did not. This can happen if backup node i is faulty.
	// Send status message to all replicas to check if they 
	// received this view change message from node i.
	// If I received a verifiable vc message from other replicas then 
	// include the view change message. If
	//  receive f+1 acks from different replicas with 
	// same digest d then include this view change message from 
	// node i. Otherwise, new primary is faulty.
      }
    }
  }

  if (mvcount < nvcount) {
    return false;
  }

  // Now I received all the view change messages that primary received and 
  // used to create new view message.
  // Now let us verify if new view message created by the new primary is 
  // correct.

  // 2. First, check the last stable checkpoint.
  int cid = nv->which_picked(nv->min());
  View_change* vc = vcs[cid].vc; 
  min = vc->last_stable();
  if (min != nv->min())
    return false;
  
  Digest d;
  if (!vc->ckpt(d,min))
    return false;

  int n_le = 1;
  int n_proofs = 1;

  // Search view-changes for proofs
  Digest dd;
  for (int i=0; i < vcs.size(); i++) {
    if (i != cid && vcs[i].vc && nv->view_change(i)) {
      if (vcs[i].vc->ckpt(dd, min) && dd == d) 
	n_proofs++;
      if (vcs[i].vc->last_stable() <= min)
	n_le++;
    }
  }

   if (n_proofs < node->f()+1 || n_le < node->n_f()) 
    return false;

  chosen_last_stable = min;
  base = min+1;

  
  // Checkpoint is correct. Check the value of nv->max(): this value
  // is correct if there exist 2f+1 view change messages that do not
  // propose any pre-prepared or prepared request with sequence number
  // greater than or equal to nv->max().
  int n_lt = 0;
  for (int i=0; i < vcs.size(); i++) {
    View_change* vc = vcs[i].vc;
    if (vc == 0 || !nv->view_change(i)) 
      continue;

    if (vc->max_seqno() < nv->max())
      n_lt++;
  }

  if (n_lt < node->n_f())
    return false;

  // nv->max() is correct. Check requests.
  max = nv->max();
  if (base == nv->max()) {
    is_complete = true;
    return true;
  }

  
  // 3. Check max_locally committed
  n_le = 1;
  n_proofs = 1;
  Digest rhd,chd;
  Seqno mlc = nv->max_committed();
  cid = nv->which_picked(mlc);
  vc = vcs[cid].vc;
  if (!vc->req(mlc,d,rhd)) return false;

  for (int i=0; i < vcs.size(); i++) {
    if (i != cid && nv->view_change(i) && vcs[i].vc) {
      if ((vcs[i].vc->last_executed() >= mlc) && vcs[i].vc->req(mlc, dd, chd) 
	  && (chd == rhd)) { 
	n_proofs++;
	n_le++;
      }
      if (vcs[i].vc->last_executed() < mlc)
	n_le++;
    }
  }

  if (n_proofs < node->f()+1 || n_le < node->n_f())  
    return false;

  chosen_max_committed = mlc;

   // 4. Check requests 
  reqs._enlarge_by(max-min-1);
  comp_reqs._enlarge_by(max-min-1);
  for (int i=0; i < comp_reqs.size(); i++) {
    comp_reqs[i] = -1;
  }
  
  for (int j=0; j < vcs.size(); j++) {
    if (nv->view_change(j)) {
      summarize_reqs(vcs[j].vc);
    }
  }
  check_complete();
  
  // 4.1 check last complete
  if (last_complete != nv->last_complete()) return false;
  
  // 4.2 check requests with complete proofs
  for (Seqno i= min+1; i <= last_complete; i++) {
    // Check if new primary picked the right id
    if (nv->which_picked(i) != reqs[i-base][comp_reqs[i-base]].id) return false;
  }

  // 4.3 Check requests without complete proofs
  //     these requests should be pointing to 
  //     new primary
  for (Seqno i=last_complete+1; i <max; i++) {
    if (nv->which_picked(i) != nv->id()) return false;
  }

  is_complete = true;        
  return true;
  
}      

void NV_info::mark_stable(Seqno ls) {
  // TO DO : Trim the data structure
}

