#ifndef _Rep_info_h
#define _Rep_info_h 1

#include <sys/time.h>
#include "types.h"
#include "Time.h"
#include "Digest.h"
#include "Array.h"
#include "Reply.h"
#include "State_defs.h"

class Req_queue;

class Rep_info {
  //
  // Holds the last replies sent to each principal.
  //
public:
  Rep_info(int nps);
  // Effects: Creates a new object that stores data for "nps"
  // principals.

  ~Rep_info();

  int num_p() const;
  // Return number of clients in the system

  Request_id req_id(int pid);

  // Returns request id of the last response sent to pid
  Digest &digest(int pid);
  // Requires: "pid" is a valid principal identifier.
  // Effects: Returns a reference to the digest of the last reply
  // value sent to pid.

  Reply* reply(int pid);
  // Requires: "pid" is a valid principal identifier.
  // Effects: Returns a pointer to the last reply value sent to "pid"
  // or 0 if no such reply was sent.

  bool add(int pid, Reply *r); 
private:
  int nps;
  Array<Reply*> reps; // Array of replies indexed by principal id.
};

inline int Rep_info::num_p() const { return nps;}
	
inline Digest& Rep_info::digest(int pid) {
  return reps[pid]->digest();
}

inline Reply* Rep_info::reply(int pid) {
  return reps[pid];
}

#endif // _Rep_info_h
