#ifndef _Req_cache_h
#define _Req_cache_h 1

#include <sys/time.h>
#include "types.h"
#include "Time.h"
#include "Digest.h"
#include "Array.h"
#include "Order_request.h"

class Req_queue;

class Req_cache {
  //
  // Holds the last replies sent to each principal.
  //
public:
  Req_cache(int nps);
  // Effects: Creates a new object that stores data for "nps"
  // principals.

  ~Req_cache();

  int num_p() const;
  // Return number of clients in the system

  Request_id req_id(int pid);

  // Returns request id of the last response sent to pid
  Digest &digest(int pid);
  // Requires: "pid" is a valid principal identifier.
  // Effects: Returns a reference to the digest of the last reply
  // value sent to pid.

  Order_request* request(int pid);
  // Requires: "pid" is a valid principal identifier.
  // Effects: Returns a pointer to the last reply value sent to "pid"
  // or 0 if no such reply was sent.

  bool add(int pid, Request_id id, Order_request *r); 
private:
  int nps;
  Array<Order_request*> reps; // Array of replies indexed by principal id.
  Array<Request_id> rid; // Array of replies indexed by principal id.
};

inline int Req_cache::num_p() const { return nps;}
	
inline Digest& Req_cache::digest(int pid) {
  return reps[pid]->request_history_digest();
}

inline Order_request* Req_cache::request(int pid) {
  return reps[pid];
}

#endif // _Req_cache_h
