#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include "Request_history.h"
#include "Array.t"
DSum* DSum::M = 0;

Req_history_log:: ~Req_history_log(){}

void Rh_entry::insert(int s, int client_id, Request_id r_id, Digest &d, char *ndet, int ndet_size, Seqno bsnum) {
  n=s;
  cid=client_id;
  rid=r_id;
  req_d = d;
  non_det_size = ndet_size;
  batch_snum = bsnum;
  if (ndet_size) {
     non_det = (char *)malloc(ndet_size);
     memcpy(non_det,ndet,ndet_size);
  }
}

void Rh_entry::digest() {
  // MD5(i, last modification seqno, (data,size)
  MD5_CTX ctx;
  MD5Init(&ctx);
  MD5Update(&ctx, (char*)&rh_d.sum,DSum::nbytes);
  MD5Final(d_h.udigest(), &ctx);
}

Req_history_log::Req_history_log(int max_out) : rh(Rh_entry(),max_out) {
  max_entries = max_out;
  last_snum = 0;
  stable_seq_num = 0;
  max_snum = 0; 

  // The random modulus for computing sums in AdHASH.
  DSum::M = new DSum;
  mpn_set_str(DSum::M->sum,
    (unsigned char*)"d2a10a09a80bc599b4d60bbec06c05d5e9f9c369954940145b63a1e2",
              DSum::nbytes, 16);

  if (sizeof(Digest)%sizeof(mp_limb_t) != 0)
    th_fail("Invalid assumption: sizeof(Digest)%sizeof(mp_limb_t)");


}

bool Req_history_log::within_range(Seqno s) {
  return ((s > stable_seq_num) && (s <= stable_seq_num+max_entries));
}

Seqno Req_history_log::update_history_digest() {
  
  // Compute request history digests of all the 
  // requests that do not have holes before their request
  // in the history log
  Seqno s = max_snum;
  //fprintf(stderr,"Update history last_snum = %lld max_snum = %lld \n", last_snum,max_snum);
  for (Seqno i=last_snum+1; i <= (int) s; i++) {
    if (rh[i%max_entries].n != -1) {
      rh[i%max_entries].rh_d=incremental_history_digest(i,rh[i%max_entries].req_d);
      rh[i%max_entries].digest();
      last_snum = i;
      //fprintf(stderr,"NO Hole for i = %lld last_snum = %lld max_snum = %lld \n", i,last_snum,max_snum);
    }
    else {
      // This can happen when there are holes 
      //fprintf(stderr,"Hole for i = %lld last_snum = %lld max_snum = %lld \n", i,last_snum,max_snum);
      break;

      //  Following claim is not true : 
      //  It is already garbage collected
      // th_assert(0,"Cannot compute request history digest \n");
    }
  }

  return last_snum;
}

// This should always be called sequentially in terms of s
DSum& Req_history_log::incremental_history_digest(Seqno s, Digest &reqd) {
  th_assert(s == last_snum+1, " Cannot compute incremental history digest. \n");
  Digest d;
  digest(d,s,reqd);

  // Update last_dsum
  last_dsum.add(d);
  return last_dsum;
}
   
void Req_history_log::digest(Digest &d, Seqno s, Digest &reqd) {
 // Compute digest for request s:
  MD5_CTX ctx;
  MD5Init(&ctx);
  MD5Update(&ctx, (char*)&s, sizeof(Seqno));
  // Request digest of rid, cid, client operation
  MD5Update(&ctx, (char*)&reqd, sizeof(Digest)); 
  // Request digest of rid, cid, client operation
  MD5Update(&ctx, (char*)rh[s%max_entries].non_det, rh[s%max_entries].non_det_size);
  MD5Final(d.udigest(), &ctx); 
}
							
bool Req_history_log::add_request(Seqno s, int cid, Request_id r_id, Digest &d, char *non_det, int non_det_size, Seqno batch_snum) {
  if (within_range(s)) {
    
    rh[s%max_entries].insert(s,cid,r_id,d,non_det,non_det_size,batch_snum);
    
    // Maximum sequence number 
    if (max_snum < s) {
      max_snum = s;    
    }
   return true;
  }
  else {
    return false;
  }
}

bool Req_history_log::comp_history_digest(Seqno s, Digest *d) {
  if (last_snum >= s) {
    Seqno bsnum = rh[s%max_entries].batch_snum;
    // if this request is batched then send the history of the
    // of the last request in the batch
    if (bsnum != Seqno_max) {
      *d=rh[bsnum%max_entries].d_h;
    }
    else{
      *d=rh[s%max_entries].d_h;
    }
    return true;
  }
  else {
    // There are holes before the request with sequence number s
    
    // Try updating history digest
    if (s <= update_history_digest()) {
      Seqno bsnum = rh[s%max_entries].batch_snum;
      // if this request is batched then send the history of the
      // of the last request in the batch
      if (bsnum != Seqno_max) {
	*d=rh[bsnum%max_entries].d_h;
      }
      else {
	*d=rh[s%max_entries].d_h;
      }
      
      return true;
    }
    else {
      return false;
    }
  }
}
    
bool Req_history_log::mark_stable(Seqno s) {
  if (last_snum < s) {
    // Wait until holes are filled
    return false;
  }
  else {
    //fprintf(stderr,"Add request: for s = %lld last_snum = %lld max_snum = %lld \n", s,last_snum,max_snum);
    stable_rh_d = rh[s%max_entries].rh_d; 
    for (long long i=stable_seq_num+1;i<s;i++) {
      rh[i%max_entries].clear();
    }
    stable_seq_num=s;
    //fprintf(stderr,"Add request: for s = %lld last_snum = %lld max_snum = %lld \n", s,last_snum,max_snum);
    return true;
  }
}
