#ifndef _Request_history_h
#define _Request_history_h 1

#include "Array.h"
#include "MD5.h"
#include "Digest.h"
#include "types.h"
#include "Request.h"
#include "DSum.h"

class Rh_entry {
public:
  inline Rh_entry() {n=-1; cid =-1; rid = -1;batch_snum=Seqno_max;non_det=0;non_det_size=-1;}
  inline Rh_entry(int s, int c, int r, Digest &d) { n=s; cid=c; rid = r; req_d = d; batch_snum=-1;} 
  void insert(int s, int client_id, Request_id r_id, Digest &d, char *ndet, int ndet_size, Seqno bsnum);
  inline ~Rh_entry() { clear();}
  void clear() {n=-1; cid =-1; rid = -1;batch_snum=Seqno_max;free(non_det); non_det=0;non_det_size=-1;}
  void digest();

  Seqno n;           // Order : seqno  
  Seqno batch_snum;  // Max sequence number of the batch that this request is part of
  int cid;           // client id
  int rid;           // request id 
  Digest req_d;      // Request digest
  DSum rh_d;         // Request history digest until this request
  Digest d_h;        // Digest of (rh_d) to reduce size from 256 bits to 128 bits
                     // We are paying more computational cost to save b/w ??
  char *non_det;
  int non_det_size; 
};

class Req_history_log {
  
 public:
  // Overview : This is a request history log with request entries ordered
  // by the assigned sequence numbers.
 
  Req_history_log(int max_out);
  // Effects: Creates an empty table.

  ~Req_history_log();
  // Effects: Deletes table 

  bool add_request(Seqno s, int cid, Request_id r_id, Digest &d, char *non_det, 
		   int non_det_size, Seqno batch_snum);
  // Creates an entry for the request in the request history log at seq no s
  // Returns true if all previous entries are filled and history digest is 
  // calculated. Otherwise returns false

  DSum& incremental_history_digest(Seqno s, Digest &reqd);
  // Compute history digest incrementally

  bool comp_history_digest(Seqno s, Digest *rh_d);
  // Returns true if request history digest of request with seq. num "s" can be 
  // copied to rh_d. Otherwise returns false

  bool request_digest(Seqno s, Digest *d);
  // Request digest with seq no s 

  bool is_batch(Seqno s);
  // Checks if request s is batched

  Seqno batch_snum(Seqno s);
  // returns the max sequence number of the batch
  
  bool within_range(Seqno s);
  // Check sequence number s is in the valid range

  Seqno update_history_digest();
  // Update history digest and return the maximum sequence number with no holes
 
  bool mark_stable(Seqno s);
  // Discards entries before s in the log and updates stable_seq_num to s
  // and copies request history digest of s to stable_rh_d;

  Seqno last_complete();
  // returns highest sequence number in the request history without any holes
  // prior to the request

  void digest(Digest &d, Seqno s, Digest &reqd);

 private:
  Array<Rh_entry> rh;
  Seqno last_snum;      // Sequence number of the request for which last history digest
                      // was computed
  DSum last_dsum;     // Adhash sum of the history

  Seqno max_snum;       // Max sequence number that is inserted
 
  Seqno stable_seq_num; // Stable sequence number
  DSum stable_rh_d;   // Stable request history digest
  int max_entries;      // Maximum entries in the log

};

inline Seqno Req_history_log::last_complete() {return last_snum;}  

inline bool Req_history_log::is_batch(Seqno s) { 
  return ((s>stable_seq_num) && (s<max_snum) && (rh[s%max_entries].batch_snum != Seqno_max));
}

inline Seqno Req_history_log::batch_snum(Seqno s) {
  return rh[s%max_entries].batch_snum;
}

inline bool Req_history_log::request_digest(Seqno s, Digest *d) {
  if (s > stable_seq_num && s <= max_snum) {
     *d = rh[s%max_entries].req_d;
     return true;
  }
  else {
     return false;
  } 
}
#endif
