#include "Request.h"
#include "Replica.h"

#include "Array.t"
#include "bhash.t"
#include "buckets.t"

Request_map::Request_map() : reqs(max_out), unordered() {
  max_entries = max_out;
}


Request_map::~Request_map() {
  MapGenerator<Digest,R_entry*> g(reqs);
  Digest d;
  R_entry* bre;
  while (g.get(d, bre)) {
    delete bre;
  }
  reqs.clear();
  unordered.clear();
} 

bool Request_map::is_ordered(Request *r) {
  Digest &rd = r->digest();
  R_entry* bre;
  return reqs.find(rd, bre);
}

// Remember that if request (req) is supplied to this function then it will not be removed from the 
// unordered queue in this routine. In such a case, caller should explicitly
// remove (as done by the primary in Order_request::Order_request
 
bool Request_map::order_request(Request *req, int cid, Request_id rid, Digest &d, View v, Seqno n) {  
  Request *r = (req) ? (req) : unordered.first_client(cid);
  R_entry* bre;

  // First check if request is present in the Map
  if (reqs.find(d, bre)) {
    // There is an entry in the map
    if (bre->r) {
      // It has full request 
      return true;
    }
    else {
      // Check if full request is present in the queue
      if (r && (rid == r->request_id()) && (d ==  r->digest())) {
        bre->r = r;
        
        // If request is read from unordered then remove it from unordered
        // else if request is supplied by the primary, it would have been 
        // already removed from the unordered request queue 

      //  BUG : We need not remove it here
      //  It gets removed when we append a new request
      // if (!req) remove_unordered(cid,rid);
        return true;
      }
      else {
        // No full request
        return false;
      }
    }
  } else {

    // No entry in reqs for rd
    // Create an entry in the map

    bre = new R_entry;
    bre->rd = d;
    bre->n = n;
    bre->v = v;
    reqs.add(d, bre);

    // Check in the unordered queue for the request
    if (r && (rid == r->request_id()) && (d ==  r->digest()))  {
      bre->r = r;
      // If request is read from unordered then remove it from unordered
      // else if request is supplied by the primary, it would have been 
      // already removed from the unordered request queue 
      //  BUG : We need not remove it here
      //  It gets removed when we append a new request
      // if (!req) remove_unordered(cid,rid);
      return true;
    }
    else {
      // No full request
      return false;
    }
  }
}


bool Request_map::add_request(Request* r) {
  th_assert(r->size() > Request::big_req_thresh && !r->is_read_only(),
	    "Invalid Argument");
  R_entry* bre;
  Digest rd = r->digest();
  if (reqs.find(rd, bre)) {
    // It is already present in the map
    // Add the request if it is not there already
    if (bre->r == 0) {
      bre->r = r;
      return true;
    }
    else {
      return false;
    }
  }
  else {
    // The request is not yet ordered
    return unordered.append(r);
  }
}


Request* Request_map::lookup(Digest& rd) {
  R_entry* bre;
  if (reqs.find(rd, bre)) {
    return bre->r;
  }
     
  return 0;
}


void Request_map::mark_stable(Seqno ls) {
  if (reqs.size() > 0) {
    MapGenerator<Digest,R_entry*> g(reqs);
    Digest d;
    R_entry* bre;
    while (g.get(d, bre)) {
      if (bre->n <= ls && bre->v >= 0) {
        //remove_unordered(bre->r->client_id(),bre->r->request_id());
	delete bre;
	g.remove();
      }
    }
  }
}


void Request_map::view_change(View v) {
  if (reqs.size() > 0) {
    MapGenerator<Digest,R_entry*> g(reqs);
    Digest d;
    R_entry* bre;
    while (g.get(d, bre)) {
      if (bre->v < v) {
	delete bre;
	g.remove();
      }
    }
  }
}
