#ifndef _Request_map_h
#define _Request_map_h 1

#include "types.h"
#include "Digest.h"
#include "Array.h"
#include "map.h"
#include "Request.h"
#include "Req_queue.h"

class R_entry {
public:
  inline R_entry() : r(0), n(-1), v(-1) {}
  inline ~R_entry() { delete r; }

  Digest rd;            // Request's digest
  Request *r;           // Request or 0 is request not received
  Seqno n;              // Order : seqno  
  View v;               // View : view
};


class Request_map {
  //
  // Overview: Table that holds entries for big requests. The entries
  // contain the requests digest and a pointer to the request or if
  // the request is not cached a list of pre-prepare messages the
  // reference the request. These entries are used to match
  // pre-prepares with their big requests. (Big requests are those
  // whose size is greater than Request::big_req_thresh.)
  //

public:
  Request_map();
  // Effects: Creates an empty table.

  ~Request_map();
  // Effects: Deletes table and any requests it references.

  bool add_request(Request* r);
  // Requires: r->size() > Request::big_req_thresh & verified == r->verify() 
  // Adds request to the map 

  Request* lookup(Digest& rd);
  // Effects: Returns the request in this with digest "rd" or 0 if
  // there is no such request.
  
  void remove_unordered(int cid, int rid);
  // remove request from the unordered array

  bool order_request(Request *req, int cid, Request_id rid, Digest& rd, View v, Seqno s);
  // Assign seq_num and view v to request with digest "rd"
  // Returns false if there is no full request with digest rd
  // Remember that if request (req) is supplied to this function then it will not be removed from the
  // unordered queue in this routine. In such a case, caller should explicitly
  // remove (as done by the primary in Order_request::Order_request)

  int pending();
  // Returns the number of requests that are unordered 

  bool is_ordered(Request *r);
  // Request true if request is ordered

  void mark_stable(Seqno ls);
  // Effects: Discards entries that were only refrred to by
  // pre-prepares that were discarded due to checkpoint "ls" becoming
  // stable.

  void view_change(View v);
  // Effects: Discards entries that were only referred to by
  // pre-prepares that were discarded due to view changing to view "v".

  int unordered_size();
  // Returns number of requests that are unordered 

  Request* unordered_remove();
  // Returns number of requests that are unordered 
private:

  Map<Digest,R_entry*> reqs;
  Req_queue unordered; // Last unordered request
  int max_entries;  // Maximum number of entries allowed in the table.
};

inline int Request_map::unordered_size() {return unordered.size();}
inline Request* Request_map::unordered_remove() {return unordered.remove();}
inline void Request_map::remove_unordered(int cid, int rid) {
  unordered.remove(cid,rid);
}

inline int Request_map::pending() {
  return unordered.size();
}

#endif // _Request_map_h
