#include <string.h>

#include "th_assert.h"
#include "Message_tags.h"
#include "Status.h"
#include "Node.h"
#include "Principal.h"
 
Status::Status(View v, Seqno ls, short t) : 
  Message(Status_tag, Max_message_size) {
  rep().v = v;        
  rep().ls = ls;
  rep().type = t;
  rep().id = node->id();
  authenticate();
}

Status::Status(View v, Seqno ls, Digest d, short t) :
  Message(Status_tag, Max_message_size) {
  rep().v = v;
  rep().ls = ls;
  rep().type = t;
  rep().reqd = d;
  rep().id = node->id();

  authenticate();
}

void Status::authenticate() {
  int old_size = sizeof(Status_rep);
  
  set_size(old_size+node->auth_size());
  // TO DO : Just create MAC for the receiver principle 
  // (random node other than principal) and the 
  // principal 
  node->gen_auth_out(contents(), old_size);

  trim();
}


bool Status::verify() {
  if (!node->is_replica(id()) || id() == node->id() || view() < 0)
    return false;

  // Check size and authenticator
  int old_size = sizeof(Status_rep);

  if (size() - old_size < node->auth_size(id()) || 
      !node->verify_auth_in(id(), contents(), old_size))
    return false;
  
  return true;
}


bool Status::convert(Message *m1, Status  *&m2) {
  if (!m1->has_tag(Status_tag, sizeof(Status_rep)))
    return false;

  m1->trim();
  m2 = (Status*)m1;
  return true;
}

