#ifndef _Status_h
#define _Status_h 1

#include "types.h"
#include "Bitmap.h"
#include "Message.h"
#include "Node.h"
#include "Digest.h"

#define GET_REQUEST_TYPE 1
#define GET_HOLE_TYPE 2

// 
// Status messages have the following format:
//

// TODO: these messages should have some kind of sequence number
// to prevent replays to be used to cause denial of service.
// TODO : Send all the holes/missing messages in a bitmap to reduce overhead
struct Status_rep : public Message_rep {
  View v;                // Replica's current view
  Seqno ls;              // seqno of the hole/missing message
  short type;            // Missing message or order request message
  Digest reqd;           // request digest of the missing request
  int id;                // id of the replica that generated the message.
};


class Status : public Message {
  // 
  //  Status messages
  //
public:
  Status(View v, Seqno ls, short type);
  // Effects: Creates a new unauthenticated Status message.  "v"
  
  Status(View v, Seqno ls, Digest reqd,  short type);

  void authenticate();
  // Effects: Authenticates message.

  int id() const;
  // Effects: Fetches the identifier of the replica from the message.

  View view() const;
  // Effects: Returns the principal id()'s view in the message.

  short type() const;
  // Effects: Returns the type : hole or missing request 

  Seqno snum() const;
  // Return the seqno of missing hole/request

  bool verify();
  // Effects: Verifies if the message has a valid authenticator from
  // id() and whether it is self-consistent.

  static bool convert(Message *m1, Status  *&m2);
  // Convert Message m1  to status message m2 

  Digest& Status::digest() const;
  // Return digest of the missing request 

private:
  Status_rep& rep() const;
  // Effects: Casts "msg" to a Status_rep&

};


inline Status_rep& Status::rep() const { 
  th_assert(ALIGNED(msg), "Improperly aligned pointer");
  return *((Status_rep*)msg); 
}


inline int Status::id() const { return rep().id; }

inline View Status::view() const { return rep().v; }

inline Seqno Status::snum() const { return rep().ls; }

inline short Status::type() const { return rep().type; }

inline Digest& Status::digest() const { return rep().reqd; }

#endif // _Status_h
