#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <limits.h>
#include <unistd.h>
#include <sys/mman.h>

extern "C" {
#include "gmp.h"
}

#include "th_assert.h"


//
// Sums of digests modulo a large integer
//
struct DSum {
  static const int nbits = 256;
  static const int mp_limb_bits = sizeof(mp_limb_t)*8;
  static const int nlimbs = (nbits+mp_limb_bits-1)/mp_limb_bits;
  static const int nbytes = nlimbs*sizeof(mp_limb_t);
  static DSum* M; // Modulus for sums must be at most nbits-1 long.

  mp_limb_t sum[nlimbs];
  //  char dummy[56];

  inline DSum() { bzero(sum, nbytes); }
  // Effects: Creates a new sum object with value 0

  inline DSum(DSum const & other) {
    memcpy(sum, other.sum, nbytes);
  }
  
  inline DSum& DSum::operator=(DSum const &other) {
    if (this == &other) return *this;
    memcpy(sum, other.sum, nbytes);
    return *this;
  }

};


DSum* DSum::M = 0;

int main() {
  DSum::M = new DSum;
  mpn_set_str(DSum::M->sum,
    (unsigned char*)"d2a10a09a80bc599b4d60bbec06c05d5e9f9c369954940145b63a1e2",
              DSum::nbytes, 16);

  printf("Size = %d bytes %d \n", DSum::nbytes,sizeof(mp_limb_t));
}
